@@ -25,7 +25,8 @@ module Storage.Hashed.Tree
     , FilterTree(..), restrict
 
     -- * Manipulating trees.
-    , modifyTree, updateTree, updateSubtrees, overlay ) where
+    , modifyTree, updateTree, partiallyUpdateTree, updateSubtrees, overlay
+    , addMissingHashes ) where
 
 import Prelude hiding( lookup, filter, all )
 import Storage.Hashed.AnchoredPath
@@ -35,7 +36,7 @@ import qualified Data.ByteString.Lazy.Char8 as BL
 import qualified Data.ByteString.Char8 as BS
 import qualified Data.Map as M
 
-import Data.Maybe( catMaybes )
+import Data.Maybe( catMaybes, isNothing )
 import Data.List( union, sort )
 import Control.Applicative( (<$>) )
 
@@ -158,10 +159,6 @@ expandUpdate update t_ = go (AnchoredPath []) t_
                 expanded_map = M.fromList expanded
                 tree = t { items = M.union orig_map expanded_map }
             update path tree
-          unstub (Stub s _) = s
-          unstub (SubTree t) = return t
-          isSub (File _) = False
-          isSub _ = True
 
 -- | Expand a stubbed Tree into a one with no stubs in it. You might want to
 -- filter the tree before expanding to save IO. This is the basic
@@ -171,21 +168,18 @@ expand :: (Monad m) => Tree m -> m (Tree m)
 expand = expandUpdate $ \_ -> return
 
 -- | Unfold a path in a (stubbed) Tree, such that the leaf node of the path is
--- reachable without crossing any stubs.
+-- reachable without crossing any stubs. Moreover, the leaf ought not be a Stub
+-- in the resulting Tree. A non-existent path is expanded as far as it can be.
 expandPath :: (Monad m) => Tree m -> AnchoredPath -> m (Tree m)
-expandPath t_ path_ = do expand' t_ path_
-    where expand' t (AnchoredPath [_]) = return t
-          expand' t (AnchoredPath (n:rest)) = do
+expandPath t_ path_ = expand' t_ path_
+    where expand' t (AnchoredPath []) = return t
+          expand' t (AnchoredPath (n:rest)) =
             case lookup t n of
-              (Just (Stub stub _)) ->
-                  do unstubbed <- stub
-                     amend t n rest unstubbed
-              (Just (SubTree t')) -> amend t n rest t'
-              _ -> fail $ "Descent error in expandPath: " ++ show path_
+              (Just item) | isSub item -> amend t n rest =<< unstub item
+              _ -> return t -- fail $ "Descent error in expandPath: " ++ show path_
           amend t name rest sub = do
             sub' <- expand' sub (AnchoredPath rest)
-            let orig_l = [ i | i@(n',_) <- listImmediate t, name /= n' ]
-                tree = t { items = M.insert name (SubTree sub') (items t) }
+            let tree = t { items = M.insert name (SubTree sub') (items t) }
             return tree
 
 class (Monad m) => FilterTree a m where
@@ -322,37 +316,40 @@ diffTrees left right =
 
 -- | Modify a Tree (by replacing, or removing or adding items).
 modifyTree :: (Monad m) => Tree m -> AnchoredPath -> Maybe (TreeItem m) -> Tree m
+modifyTree t_ p_ i_ = snd $ go t_ p_ i_
+  where fix t unmod items' = (unmod, t { items = countmap items' `seq` items'
+                                       , treeHash = if unmod then treeHash t else NoHash })
 
-modifyTree _ (AnchoredPath []) (Just (SubTree sub)) = sub
+        go t (AnchoredPath []) (Just (SubTree sub)) = (treeHash t `match` treeHash sub, sub)
 
-modifyTree t (AnchoredPath [n]) (Just item) =
-    t { items = countmap items' `seq` items'
-      , treeHash = NoHash }
-  where !items' = M.insert n item (items t)
+        go t (AnchoredPath [n]) (Just item) = fix t unmod items'
+            where !items' = M.insert n item (items t)
+                  !unmod = itemHash item `match` case lookup t n of
+                                             Nothing -> NoHash
+                                             Just i -> itemHash i
 
-modifyTree t (AnchoredPath [n]) Nothing =
-    t { items = countmap items' `seq` items'
-      , treeHash = NoHash }
-  where !items' = M.delete n (items t)
+        go t (AnchoredPath [n]) Nothing = fix t unmod items'
+            where !items' = M.delete n (items t)
+                  !unmod = isNothing $ lookup t n
 
-modifyTree t path@(AnchoredPath (n:r)) item =
-    t { items = countmap items' `seq` items'
-      , treeHash = NoHash }
-  where subtree s = modifyTree s (AnchoredPath r) item
-        !items' = M.insert n sub (items t)
-        !sub = case lookup t n of
-                Just (SubTree s) -> SubTree $! subtree s
-                Just (Stub s _) -> Stub (do x <- s
-                                            return $! subtree x) NoHash
-                Nothing -> SubTree $! subtree emptyTree
-                _ -> error $ "Modify tree at " ++ show path
+        go t path@(AnchoredPath (n:r)) item = fix t unmod items'
+            where subtree s = go s (AnchoredPath r) item
+                  !items' = M.insert n sub (items t)
+                  !sub = snd sub'
+                  !unmod = fst sub'
+                  !sub' = case lookup t n of
+                    Just (SubTree s) -> let (mod, sub) = subtree s in (mod, SubTree sub)
+                    Just (Stub s _) -> (False, Stub (do x <- s
+                                                        return $! snd $! subtree x) NoHash)
+                    Nothing -> (False, SubTree $! snd $! subtree emptyTree)
+                    _ -> error $ "Modify tree at " ++ show path
 
-modifyTree _ (AnchoredPath []) (Just (Stub _ _)) =
-    error "Bug in descent in modifyTree."
-modifyTree _ (AnchoredPath []) (Just (File _)) =
-    error "Bug in descent in modifyTree."
-modifyTree _ (AnchoredPath []) Nothing =
-    error "Bug in descent in modifyTree."
+        go _ (AnchoredPath []) (Just (Stub _ _)) =
+            error $ "BUG: Error descending in modifyTree, path = " ++ show p_
+        go _ (AnchoredPath []) (Just (File _)) =
+            error $ "BUG: Error descending in modifyTree, path = " ++ show p_
+        go _ (AnchoredPath []) Nothing =
+            error $ "BUG: Error descending in modifyTree, path = " ++ show p_
 
 countmap = M.fold (\_ i -> i + 1) 0
 
@@ -366,13 +363,22 @@ updateSubtrees fun t =
 
 -- | Does /not/ expand the tree.
 updateTree :: (Functor m, Monad m) => (TreeItem m -> m (TreeItem m)) -> Tree m -> m (Tree m)
-updateTree fun t = do
-    immediate <- mapM update $ listImmediate t
-    SubTree t' <- fun . SubTree $ t { items = M.fromList immediate
-                                    , treeHash = NoHash }
-    return t'
-  where update (k, SubTree tree) = (\new -> (k, SubTree new)) <$> updateTree fun tree
-        update (k, item) = (\new -> (k, new)) <$> fun item
+updateTree fun t = partiallyUpdateTree fun (\_ _ -> True) t
+
+-- | Does /not/ expand the tree.
+partiallyUpdateTree :: (Functor m, Monad m) => (TreeItem m -> m (TreeItem m))
+                       -> (AnchoredPath -> TreeItem m -> Bool) -> Tree m -> m (Tree m)
+partiallyUpdateTree fun pred t' = go (AnchoredPath []) t'
+  where go path t = do
+          items' <- M.fromList <$> mapM (maybeupdate path) (listImmediate t)
+          SubTree t' <- fun . SubTree $ t { items = items'
+                                          , treeHash = NoHash }
+          return t'
+        maybeupdate path (k, item) = case pred (path `appendPath` k) item of
+          True -> update (path `appendPath` k) (k, item)
+          False -> return (k, item)
+        update path (k, SubTree tree) = (\new -> (k, SubTree new)) <$> go path tree
+        update    _ (k, item) = (\new -> (k, new)) <$> fun item
 
 -- | Lay one tree over another. The resulting Tree will look like the base (1st
 -- parameter) Tree, although any items also present in the overlay Tree will be
@@ -394,3 +400,23 @@ overlay base over = Tree { items = M.fromList immediate
                                                                    return $ overlay b' o') NoHash
                     (Just x, _) -> x
                     (_, _) -> error $ "Unexpected case in overlay at get " ++ show n ++ "."
+
+addMissingHashes :: (Monad m, Functor m) => (TreeItem m -> m Hash) -> Tree m -> m (Tree m)
+addMissingHashes make = updateTree update -- use partiallyUpdateTree here
+    where update (SubTree t) = make (SubTree t) >>= \x -> return $ SubTree (t { treeHash = x })
+          update (File blob@(Blob con NoHash)) =
+              do hash <- make $ File blob
+                 return $ File (Blob con hash)
+          update (Stub s NoHash) = update . SubTree =<< s
+          update x = return x
+
+------ Private utilities shared among multiple functions. --------
+
+unstub :: (Monad m) => TreeItem m -> m (Tree m)
+unstub (Stub s _) = s
+unstub (SubTree s) = return s
+
+isSub :: TreeItem m -> Bool
+isSub (File _) = False
+isSub _ = True
+
